package zio

import org.scalacheck._
import org.specs2.ScalaCheck
import org.specs2.execute.Result

import scala.collection.mutable
import scala.util.Try
import zio.Cause.{ die, fail, interrupt, Both }
import zio.duration._
import zio.test.mock.MockClock

class IOSpec(implicit ee: org.specs2.concurrent.ExecutionEnv) extends TestRuntime with GenIO with ScalaCheck {
  import Prop.forAll

  def is = "IOSpec".title ^ s2"""
   Generate a list of String and a f: String => Task[Int]:
      `IO.foreach` returns the list of results. $t1
   Create a list of Strings and pass an f: String => IO[String, Int]:
      `IO.foreach` both evaluates effects and returns the list of Ints in the same order. $t2
   Create a list of String and pass an f: String => IO[String, Int]:
      `IO.foreach` fails with a NumberFormatException exception. $t3
   Create a list of Strings and pass an f: String => IO[String, Int]:
      `IO.foreachPar` returns the list of Ints in the same order. $t4
   Create an integer and an f: Int => String:
      `IO.bimap(f, identity)` maps an IO[Int, String] into an IO[String, String]. $t5
   Create a list of Ints and map with IO.point:
      `IO.collectAllPar` returns the list of Ints in the same order. $t6
   Create a list of Ints and map with IO.point:
      `IO.forkAll` returns the list of Ints in the same order. $t7
   Create a list of Strings and pass an f: String => UIO[Int]:
      `IO.collectAllParN` returns the list of Ints in the same order. $t8
   Create a list of Ints and pass an f: Int => UIO[Int]:
      `IO.foreachParN` returns the list of created Strings in the appropriate order. $t9
   Create a list of Ints:
      `IO.foldLeft` with a successful step function sums the list properly. $t10
   Create a non-empty list of Ints:
      `IO.foldLeft` with a failing step function returns a failed IO. $t11
   Check done lifts exit result into IO. $testDone
   Check `catchSomeCause` catches matching cause $testCatchSomeCauseMatch
   Check `catchSomeCause` halts if cause doesn't match $testCatchSomeCauseNoMatch
   Check `when` executes correct branch only. $testWhen
   Check `whenM` executes condition effect and correct branch. $testWhenM
   Check `whenCase` executes correct branch only. $testWhenCase
   Check `whenCaseM` executes condition effect and correct branch. $testWhenCaseM
   Check `unsandbox` unwraps exception. $testUnsandbox
   Check `supervise` returns same value as IO.supervise. $testSupervise
   Check `flatten` method on IO[E, IO[E, String] returns the same IO[E, String] as `IO.flatten` does. $testFlatten
   Check `absolve` method on IO[E, Either[E, A]] returns the same IO[E, Either[E, String]] as `IO.absolve` does. $testAbsolve
   Check non-`memoize`d IO[E, A] returns new instances on repeated calls due to referential transparency. $testNonMemoizationRT
   Check `memoize` method on IO[E, A] returns the same instance on repeated calls. $testMemoization
   Check `cached` method on IO[E, A] returns new instances after duration. $testCached
   Check `raceAll` method returns the same IO[E, A] as `IO.raceAll` does. $testRaceAll
   Check `firstSuccessOf` method returns the same IO[E, A] as `IO.firstSuccessOf` does. $testfirstSuccessOf
   Check `zipPar` method does not swallow exit causes of loser. $testZipParInterupt
   Check `zipPar` method does not report failure when interrupting loser after it succeeded. $testZipParSucceed
   Check `orElse` method does not recover from defects. $testOrElseDefectHandling
   Check `eventually` method succeeds eventually. $testEventually
   Check `some` method extracts the value from Some. $testSomeOnSomeOption
   Check `some` method fails on None. $testSomeOnNoneOption
   Check `some` method fails when given an exception. $testSomeOnException
   Check `someOrFail` method extracts the optional value. $testSomeOrFailExtractOptionalValue
   Check `someOrFail` method fails when given a None. $testSomeOrFailWithNone
   Check `someOrFailException` method extracts the optional value. $testSomeOrFailExceptionOnOptionalValue
   Check `someOrFailException` method fails when given a None. $testSomeOrFailExceptionOnEmptyValue
   Check `none` method extracts the value from Some. $testNoneOnSomeOption
   Check `none` method fails on None. $testNoneOnNoneOption
   Check `none` method fails when given an exception. $testNoneOnException
   Check `flattenErrorOption` method fails when given Some error. $testFlattenErrorOptionOnSomeError
   Check `flattenErrorOption` method fails with Default when given None error. $testFlattenErrorOptionOnNoneError
   Check `flattenErrorOption` method succeeds when given a value. $testFlattenErrorOptionOnSomeValue
   Check `optional` method fails when given Some error. $testOptionalOnSomeError
   Check `optional` method succeeds with None given None error. $testOptionalOnNoneError
   Check `optional` method succeeds with Some given a value. $testOptionalOnSomeValue
   Check `right` method extracts the Right value. $testRightOnRightValue
   Check `right` method fails with `None` when given a Left value. $testRightOnLeftValue
   Check `right` method fails with `Some(exception)` when given an exception. $testRightOnException
   Check `rightOrFail` method extracts the Right value. $testRightOrFailExtractsRightValue
   Check `rightOrFail` method fails when given a Left. $testRightOrFailWithLeft
   Check `rightOrFailException` method extracts the Right value. $testRightOrFailExceptionOnRightValue
   Check `rightOrFailException` method fails when given a Left. $testRightOrFailExceptionOnLeftValue
   Check `left` method extracts the Left value. $testLeftOnLeftValue
   Check `left` method fails with `None` when given a Right value. $testLeftOnRightValue
   Check `left` method fails with `Some(exception)` when given an exception. $testLeftOnException
   Check `leftOrFail` method extracts the Left value. $testLeftOrFailExtractsLeftValue
   Check `leftOrFail` method fails when given a Right. $testLeftOrFailWithRight
   Check `leftOrFailException` method extracts the Left value. $testLeftOrFailExceptionOnLeftValue
   Check `leftOrFailException` method fails when given a Right. $testLeftOrFailExceptionOnRightValue
   Check `head` method extracts the head of a non-empty list. $testHeadOnNonEmptyList
   Check `head` method fails with `None` when given an empty list. $testHeadOnEmptyList
   Check `head` method fails with `Some(exception)` when given an exception. $testHeadOnException
   Check `replicate` method returns empty list when given non positive number. $testReplicateNonPositiveNumber
   Check `replicate` method returns list of the same effect. $testReplicate
   Check uncurried `bracket`. $testUncurriedBracket
   Check uncurried `bracket_`. $testUncurriedBracket_
   Check uncurried `bracketExit`. $testUncurriedBracketExit
   Check `bracketExit` error handling. $testBracketExitErrorHandling
   Check `foreach_` runs effects in order. $testForeach_Order
   Check `foreach_` can be run twice. $testForeach_Twice
   Check `foreachPar_` runs all effects. $testForeachPar_Full
   Check `foreachParN_` runs all effects. $testForeachParN_Full
   Check `filterOrElse` returns checked failure from held value $testFilterOrElse
   Check `filterOrElse_` returns checked failure ignoring value $testFilterOrElse_
   Check `filterOrFail` returns failure ignoring value $testFilterOrFail
   Check `collect` returns failure ignoring value $testCollect
   Check `collectM` returns failure ignoring value $testCollectM
   Check `reject` returns failure ignoring value $testReject
   Check `rejectM` returns failure ignoring value $testRejectM
   Check `foreachParN` works on large lists $testForeachParN_Threads
   Check `foreachParN` runs effects in parallel $testForeachParN_Parallel
   Check `foreachParN` propogates error $testForeachParN_Error
   Check `foreachParN` interrupts effects on first failure $testForeachParN_Interruption
    """

  def functionIOGen: Gen[String => Task[Int]] =
    Gen.function1[String, Task[Int]](genSuccess[Throwable, Int])

  def listGen: Gen[List[String]] =
    Gen.listOfN(100, Gen.alphaNumStr)

  def t1 = forAll(functionIOGen, listGen) { (f, list) =>
    val res = unsafeRun(IO.foreach(list)(f))
    res must be size 100
    res must beAnInstanceOf[List[Int]]
  }

  def t2 = {
    val list    = List("1", "2", "3")
    val effects = new mutable.ListBuffer[String]
    val res     = unsafeRun(IO.foreach(list)(x => IO.effectTotal(effects += x) *> IO.effectTotal[Int](x.toInt)))
    (effects.toList, res) must be_===((list, List(1, 2, 3)))
  }

  def t3 = {
    val list = List("1", "h", "3")
    val res  = Try(unsafeRun(IO.foreach(list)(x => IO.effectTotal[Int](x.toInt))))
    res must beAFailedTry.withThrowable[FiberFailure]
  }

  def t4 = {
    val list = List("1", "2", "3")
    val res  = unsafeRun(IO.foreachPar(list)(x => IO.effectTotal[Int](x.toInt)))
    res must be_===(List(1, 2, 3))
  }

  def t5 = forAll { (i: Int) =>
    val res = unsafeRun(IO.fail[Int](i).bimap(_.toString, identity).either)
    res must_=== Left(i.toString)
  }

  def t6 = {
    val list = List(1, 2, 3).map(IO.effectTotal[Int](_))
    val res  = unsafeRun(IO.collectAllPar(list))
    res must be_===(List(1, 2, 3))
  }

  def t7 = {
    val list = List(1, 2, 3).map(IO.effectTotal[Int](_))
    val res  = unsafeRun(IO.forkAll(list).flatMap[Any, Nothing, List[Int]](_.join))
    res must be_===(List(1, 2, 3))
  }

  def t8 = {
    val list = List(1, 2, 3).map(IO.effectTotal[Int](_))
    val res  = unsafeRun(IO.collectAllParN(2)(list))
    res must be_===(List(1, 2, 3))
  }

  def t9 = {
    val list = List(1, 2, 3)
    val res  = unsafeRun(IO.foreachParN(2)(list)(x => IO.effectTotal(x.toString)))
    res must be_===(List("1", "2", "3"))
  }

  def t10 = forAll { (l: List[Int]) =>
    unsafeRun(IO.foldLeft(l)(0)((acc, el) => IO.succeed(acc + el))) must_=== unsafeRun(IO.succeed(l.sum))
  }

  def t11 = forAll { (l: List[Int]) =>
    l.size > 0 ==>
      (unsafeRunSync(IO.foldLeft(l)(0)((_, _) => IO.fail("fail"))) must_=== unsafeRunSync(IO.fail("fail")))
  }

  private val exampleError = new Error("something went wrong")

  def testDone = {
    val error                         = exampleError
    val completed                     = Exit.succeed(1)
    val interrupted: Exit[Error, Int] = Exit.interrupt
    val terminated: Exit[Error, Int]  = Exit.die(error)
    val failed: Exit[Error, Int]      = Exit.fail(error)

    unsafeRun(IO.done(completed)) must_=== 1
    unsafeRunSync(IO.done(interrupted)) must_=== Exit.interrupt
    unsafeRunSync(IO.done(terminated)) must_=== Exit.die(error)
    unsafeRunSync(IO.done(failed)) must_=== Exit.fail(error)
  }

  private def testCatchSomeCauseMatch =
    unsafeRun {
      ZIO.interrupt.catchSomeCause {
        case c if (c.interrupted) => ZIO.succeed(true)
      }.sandbox.map(_ must_=== true)
    }

  private def testCatchSomeCauseNoMatch =
    unsafeRun {
      ZIO.interrupt.catchSomeCause {
        case c if (!c.interrupted) => ZIO.succeed(true)
      }.sandbox.either.map(_ must_=== Left(Cause.interrupt))
    }

  def testWhen =
    unsafeRun(
      for {
        effectRef <- Ref.make(0)
        _         <- effectRef.set(1).when(false)
        val1      <- effectRef.get
        _         <- effectRef.set(2).when(true)
        val2      <- effectRef.get
        failure   = new Exception("expected")
        _         <- IO.fail(failure).when(false)
        failed    <- IO.fail(failure).when(true).either
      } yield (val1 must_=== 0) and
        (val2 must_=== 2) and
        (failed must beLeft(failure))
    )

  def testWhenM =
    unsafeRun(
      for {
        effectRef      <- Ref.make(0)
        conditionRef   <- Ref.make(0)
        conditionTrue  = conditionRef.update(_ + 1).map(_ => true)
        conditionFalse = conditionRef.update(_ + 1).map(_ => false)
        _              <- effectRef.set(1).whenM(conditionFalse)
        val1           <- effectRef.get
        conditionVal1  <- conditionRef.get
        _              <- effectRef.set(2).whenM(conditionTrue)
        val2           <- effectRef.get
        conditionVal2  <- conditionRef.get
        failure        = new Exception("expected")
        _              <- IO.fail(failure).whenM(conditionFalse)
        failed         <- IO.fail(failure).whenM(conditionTrue).either
      } yield (val1 must_=== 0) and
        (conditionVal1 must_=== 1) and
        (val2 must_=== 2) and
        (conditionVal2 must_=== 2) and
        (failed must beLeft(failure))
    )

  def testWhenCase =
    unsafeRun {
      val v1: Option[Int] = None
      val v2: Option[Int] = Some(0)
      for {
        ref  <- Ref.make(false)
        _    <- ZIO.whenCase(v1) { case Some(_) => ref.set(true) }
        res1 <- ref.get
        _    <- ZIO.whenCase(v2) { case Some(_) => ref.set(true) }
        res2 <- ref.get
      } yield (res1 must_=== false) and (res2 must_=== true)
    }

  def testWhenCaseM =
    unsafeRun {
      val v1: Option[Int] = None
      val v2: Option[Int] = Some(0)
      for {
        ref  <- Ref.make(false)
        _    <- ZIO.whenCaseM(IO.succeed(v1)) { case Some(_) => ref.set(true) }
        res1 <- ref.get
        _    <- ZIO.whenCaseM(IO.succeed(v2)) { case Some(_) => ref.set(true) }
        res2 <- ref.get
      } yield (res1 must_=== false) and (res2 must_=== true)
    }

  def testUnsandbox = {
    val failure: IO[Cause[Exception], String] = IO.fail(fail(new Exception("fail")))
    val success: IO[Cause[Any], Int]          = IO.succeed(100)
    unsafeRun(for {
      message <- failure.unsandbox.foldM(e => IO.succeed(e.getMessage), _ => IO.succeed("unexpected"))
      result  <- success.unsandbox
    } yield (message must_=== "fail") and (result must_=== 100))
  }

  def testSupervise = {
    val io = IO.effectTotal("supercalifragilisticexpialadocious")
    unsafeRun(for {
      supervise1 <- io.interruptChildren
      supervise2 <- IO.interruptChildren(io)
    } yield supervise1 must ===(supervise2))
  }

  def testFlatten = forAll(Gen.alphaStr) { str =>
    unsafeRun(for {
      flatten1 <- IO.effectTotal(IO.effectTotal(str)).flatten
      flatten2 <- IO.flatten(IO.effectTotal(IO.effectTotal(str)))
    } yield flatten1 must ===(flatten2))
  }

  def testAbsolve = forAll(Gen.alphaStr) { str =>
    val ioEither: UIO[Either[Nothing, String]] = IO.succeed(Right(str))
    unsafeRun(for {
      abs1 <- ioEither.absolve
      abs2 <- IO.absolve(ioEither)
    } yield abs1 must ===(abs2))
  }

  def testNonMemoizationRT = forAll(Gen.alphaStr) { str =>
    val io: UIO[Option[String]] = IO.effectTotal(Some(str)) // using `Some` for object allocation
    unsafeRun(
      (io <*> io)
        .map(tuple => tuple._1 must not beTheSameAs (tuple._2))
    )
  }

  def testMemoization = forAll(Gen.alphaStr) { str =>
    val ioMemo: UIO[UIO[Option[String]]] = IO.effectTotal(Some(str)).memoize // using `Some` for object allocation
    unsafeRun(
      ioMemo
        .flatMap(io => io <*> io)
        .map(tuple => tuple._1 must beTheSameAs(tuple._2))
    )
  }

  def testCached = unsafeRunWith(MockClock.make(MockClock.DefaultData)) {
    def incrementAndGet(ref: Ref[Int]): UIO[Int] = ref.update(_ + 1)
    for {
      ref   <- Ref.make(0)
      cache <- incrementAndGet(ref).cached(60.minutes)
      a     <- cache
      _     <- MockClock.adjust(59.minutes)
      b     <- cache
      _     <- MockClock.adjust(1.minute)
      c     <- cache
      _     <- MockClock.adjust(59.minutes)
      d     <- cache
    } yield (a must_=== b) and (b must_!== c) and (c must_=== d)
  }

  def testRaceAll = {
    val io  = IO.effectTotal("supercalifragilisticexpialadocious")
    val ios = List.empty[UIO[String]]
    unsafeRun(for {
      race1 <- io.raceAll(ios)
      race2 <- IO.raceAll(io, ios)
    } yield race1 must ===(race2))
  }

  def testfirstSuccessOf = {
    val io  = IO.effectTotal("supercalifragilisticexpialadocious")
    val ios = List.empty[UIO[String]]
    unsafeRun(for {
      race1 <- io.firstSuccessOf(ios)
      race2 <- IO.firstSuccessOf(io, ios)
    } yield race1 must ===(race2))
  }

  def testZipParInterupt = {
    val io = ZIO.interrupt.zipPar(IO.interrupt)
    unsafeRunSync(io) must_=== Exit.Failure(Both(interrupt, interrupt))
  }

  def testZipParSucceed = {
    val io = ZIO.interrupt.zipPar(IO.succeed(1))
    unsafeRun(io.sandbox.either).left.map(_.interrupted) must_=== Left(true)
  }

  def testOrElseDefectHandling = {
    val ex = new Exception("Died")

    unsafeRun {
      for {
        plain <- (ZIO.die(ex) <> IO.unit).run
        both  <- (ZIO.halt(Cause.Both(interrupt, die(ex))) <> IO.unit).run
        thn   <- (ZIO.halt(Cause.Then(interrupt, die(ex))) <> IO.unit).run
        fail  <- (ZIO.fail(ex) <> IO.unit).run
      } yield (plain must_=== Exit.die(ex))
        .and(both must_=== Exit.die(ex))
        .and(thn must_=== Exit.die(ex))
        .and(fail must_=== Exit.succeed(()))
    }
  }

  def testEventually = {
    def effect(ref: Ref[Int]) =
      ref.get.flatMap(n => if (n < 10) ref.update(_ + 1) *> IO.fail("Ouch") else UIO.succeed(n))

    val test = for {
      ref <- Ref.make(0)
      n   <- effect(ref).eventually
    } yield n

    unsafeRun(test) must_=== 10
  }

  def testSomeOnSomeOption = {
    val task: IO[Option[Throwable], Int] = Task(Some(1)).some
    unsafeRun(task) must_=== 1
  }

  def testSomeOnNoneOption = {
    val task: IO[Option[Throwable], Int] = Task(None).some
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testSomeOnException = {
    val task: IO[Option[Throwable], Int] = Task.fail(new RuntimeException("Failed Task")).some
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testNoneOnSomeOption = {
    val task: IO[Option[Throwable], Unit] = Task(Some(1)).none
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testNoneOnNoneOption = {
    val task: IO[Option[Throwable], Unit] = Task(None).none
    unsafeRun(task) must_=== (())
  }

  def testNoneOnException = {
    val task: IO[Option[Throwable], Unit] = Task.fail(new RuntimeException("Failed Task")).none
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testFlattenErrorOptionOnSomeError = {
    val task: IO[String, Int] = IO.fail(Some("Error")).flattenErrorOption("Default")
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testFlattenErrorOptionOnNoneError = {
    val task: IO[String, Int] = IO.fail(None).flattenErrorOption("Default")
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testFlattenErrorOptionOnSomeValue = {
    val task: IO[String, Int] = IO.succeed(1).flattenErrorOption("Default")
    unsafeRun(task) must_=== 1
  }

  def testOptionalOnSomeError = {
    val task: IO[String, Option[Int]] = IO.fail(Some("Error")).optional
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testOptionalOnNoneError = {
    val task: IO[String, Option[Int]] = IO.fail(None).optional
    unsafeRun(task) must_=== None
  }

  def testOptionalOnSomeValue = {
    val task: IO[String, Option[Int]] = IO.succeed(1).optional
    unsafeRun(task) must_=== Some(1)
  }

  def testSomeOrFailWithNone = {
    val task: Task[Int] = UIO(Option.empty[Int]).someOrFail(exampleError)
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testSomeOrFailExtractOptionalValue = {
    val task: Task[Int] = UIO(Some(42)).someOrFail(exampleError)
    unsafeRun(task) must_=== 42
  }

  def testSomeOrFailExceptionOnOptionalValue = unsafeRun(ZIO.succeed(Some(42)).someOrFailException) must_=== 42

  def testSomeOrFailExceptionOnEmptyValue = {
    val task = ZIO.succeed(Option.empty[Int]).someOrFailException
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testRightOnRightValue = {
    val task = ZIO.succeed(Right("Right")).right.either
    unsafeRun(task) must_=== Right("Right")
  }

  def testRightOnLeftValue = {
    val task = ZIO.succeed(Left("Left")).right.either
    unsafeRun(task) must_=== Left(None)
  }

  def testRightOnException = {
    val task = ZIO.fail("Fail").right.either
    unsafeRun(task) must_=== Left(Some("Fail"))
  }

  def testRightOrFailExceptionOnRightValue = unsafeRun(ZIO.succeed(Right(42)).rightOrFailException) must_=== 42

  def testRightOrFailExceptionOnLeftValue = {
    val task: Task[Int] = ZIO.succeed(Left(2)).rightOrFailException
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testRightOrFailExtractsRightValue = {
    val task: Task[Int] = UIO(Right(42)).rightOrFail(exampleError)
    unsafeRun(task) must_=== 42
  }

  def testRightOrFailWithLeft = {
    val task: Task[Int] = UIO(Left(1)).rightOrFail(exampleError)
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testLeftOnLeftValue = {
    val task = ZIO.succeed(Left("Left")).left.either
    unsafeRun(task) must_=== Right("Left")
  }

  def testLeftOnRightValue = {
    val task = ZIO.succeed(Right("Right")).left.either
    unsafeRun(task) must_=== Left(None)
  }

  def testLeftOnException = {
    val task = ZIO.fail("Fail").left.either
    unsafeRun(task) must_=== Left(Some("Fail"))
  }

  def testLeftOrFailExceptionOnLeftValue = unsafeRun(ZIO.succeed(Left(42)).leftOrFailException) must_=== 42

  def testLeftOrFailExceptionOnRightValue = {
    val task: Task[Int] = ZIO.succeed(Right(2)).leftOrFailException
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testReplicateNonPositiveNumber = {
    val lst: Iterable[UIO[Int]] = ZIO.replicate(0)(ZIO.succeed(12))
    unsafeRun(ZIO.sequence(lst)) must_=== List()

    val anotherList: Iterable[UIO[Int]] = ZIO.replicate(-2)(ZIO.succeed(12))
    unsafeRun(ZIO.sequence(anotherList)) must_=== List()
  }

  def testReplicate = {
    val lst: Iterable[UIO[Int]] = ZIO.replicate(2)(ZIO.succeed(12))
    unsafeRun(ZIO.sequence(lst)) must_=== List.fill(2)(12)
  }

  def testLeftOrFailExtractsLeftValue = {
    val task: Task[Int] = UIO(Left(42)).leftOrFail(exampleError)
    unsafeRun(task) must_=== 42
  }

  def testLeftOrFailWithRight = {
    val task: Task[Int] = UIO(Right(12)).leftOrFail(exampleError)
    unsafeRun(task) must throwA[FiberFailure]
  }

  def testHeadOnNonEmptyList = {
    val task = ZIO.succeed(List(1, 2, 3)).head.either
    unsafeRun(task) must_=== Right(1)
  }

  def testHeadOnEmptyList = {
    val task = ZIO.succeed(List.empty).head.either
    unsafeRun(task) must_=== Left(None)
  }

  def testHeadOnException = {
    val task = ZIO.fail("Fail").head.either
    unsafeRun(task) must_=== Left(Some("Fail"))
  }

  def testUncurriedBracket =
    unsafeRun {
      for {
        release  <- Ref.make(false)
        result   <- ZIO.bracket(IO.succeed(42), (_: Int) => release.set(true), (a: Int) => ZIO.effectTotal(a + 1))
        released <- release.get
      } yield (result must_=== 43) and (released must_=== true)
    }

  def testUncurriedBracket_ =
    unsafeRun {
      for {
        release  <- Ref.make(false)
        result   <- IO.succeed(42).bracket_(release.set(true), ZIO.effectTotal(0))
        released <- release.get
      } yield (result must_=== 0) and (released must_=== true)
    }

  def testUncurriedBracketExit =
    unsafeRun {
      for {
        release <- Ref.make(false)
        result <- ZIO.bracketExit(
                   IO.succeed(42),
                   (_: Int, _: Exit[_, _]) => release.set(true),
                   (_: Int) => IO.succeed(0L)
                 )
        released <- release.get
      } yield (result must_=== 0L) and (released must_=== true)
    }

  def testBracketExitErrorHandling = {
    val releaseDied = new RuntimeException("release died")
    val exit: Exit[String, Int] = unsafeRunSync {
      ZIO.bracketExit[Any, String, Int, Int](
        ZIO.succeed(42),
        (_, _) => ZIO.die(releaseDied),
        _ => ZIO.fail("use failed")
      )
    }

    exit.fold[Result](
      cause => (cause.failures must_=== List("use failed")) and (cause.defects must_=== List(releaseDied)),
      value => failure(s"unexpectedly completed with value $value")
    )
  }

  object UncurriedBracketCompilesRegardlessOrderOfEAndRTypes {
    class A
    class B
    class R
    class R1 extends R
    class R2 extends R1
    class E
    class E1 extends E

    def infersEType1: ZIO[R, E, B] = {
      val acquire: ZIO[R, E, A]            = ???
      val release: A => ZIO[R, Nothing, _] = ???
      val use: A => ZIO[R, E1, B]          = ???
      ZIO.bracket(acquire, release, use)
    }

    def infersEType2: ZIO[R, E, B] = {
      val acquire: ZIO[R, E1, A]           = ???
      val release: A => ZIO[R, Nothing, _] = ???
      val use: A => ZIO[R, E, B]           = ???
      ZIO.bracket(acquire, release, use)
    }

    def infersRType1: ZIO[R2, E, B] = {
      val acquire: ZIO[R, E, A]             = ???
      val release: A => ZIO[R1, Nothing, _] = ???
      val use: A => ZIO[R2, E, B]           = ???
      ZIO.bracket(acquire, release, use)
    }

    def infersRType2: ZIO[R2, E, B] = {
      val acquire: ZIO[R2, E, A]            = ???
      val release: A => ZIO[R1, Nothing, _] = ???
      val use: A => ZIO[R, E, B]            = ???
      ZIO.bracket(acquire, release, use)
    }

    def infersRType3: ZIO[R2, E, B] = {
      val acquire: ZIO[R1, E, A]            = ???
      val release: A => ZIO[R2, Nothing, _] = ???
      val use: A => ZIO[R, E, B]            = ???
      ZIO.bracket(acquire, release, use)
    }
  }

  def testForeach_Order = {
    val as = List(1, 2, 3, 4, 5)
    val r = unsafeRun {
      for {
        ref <- Ref.make(List.empty[Int])
        _   <- ZIO.foreach_(as)(a => ref.update(_ :+ a))
        rs  <- ref.get
      } yield rs
    }
    r must_== as
  }

  def testForeach_Twice = {
    val as = List(1, 2, 3, 4, 5)
    val r = unsafeRun {
      for {
        ref <- Ref.make(0)
        zio = ZIO.foreach_(as)(a => ref.update(_ + a))
        _   <- zio
        _   <- zio
        sum <- ref.get
      } yield sum
    }
    r must_=== 30
  }

  def testForeachPar_Full = {
    val as = Seq(1, 2, 3, 4, 5)
    val r = unsafeRun {
      for {
        ref <- Ref.make(Seq.empty[Int])
        _   <- ZIO.foreachPar_(as)(a => ref.update(_ :+ a))
        rs  <- ref.get
      } yield rs
    }
    r must have length as.length
    r must containTheSameElementsAs(as)
  }

  def testForeachParN_Full = {
    val as = Seq(1, 2, 3, 4, 5)
    val r = unsafeRun {
      for {
        ref <- Ref.make(Seq.empty[Int])
        _   <- ZIO.foreachParN_(2)(as)(a => ref.update(_ :+ a))
        rs  <- ref.get
      } yield rs
    }
    r must have length as.length
    r must containTheSameElementsAs(as)
  }

  def testFilterOrElse = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.filterOrElse(_ == 0)(a => ZIO.fail(s"$a was not 0"))).sandbox.either
    ) must_=== Right(0)

    val badCase = unsafeRun(
      exactlyOnce(1)(_.filterOrElse(_ == 0)(a => ZIO.fail(s"$a was not 0"))).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("1 was not 0"))

    goodCase and badCase
  }

  def testFilterOrElse_ = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.filterOrElse_(_ == 0)(ZIO.fail("Predicate failed!"))).sandbox.either
    ) must_=== Right(0)

    val badCase = unsafeRun(
      exactlyOnce(1)(_.filterOrElse_(_ == 0)(ZIO.fail("Predicate failed!"))).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Predicate failed!"))

    goodCase and badCase
  }

  def testFilterOrFail = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.filterOrFail(_ == 0)("Predicate failed!")).sandbox.either
    ) must_=== Right(0)

    val badCase = unsafeRun(
      exactlyOnce(1)(_.filterOrFail(_ == 0)("Predicate failed!")).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Predicate failed!"))

    goodCase and badCase
  }

  def testCollect = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.collect(s"value was not 0")({ case v @ 0 => v })).sandbox.either
    ) must_=== Right(0)

    val badCase = unsafeRun(
      exactlyOnce(1)(_.collect(s"value was not 0")({ case v @ 0 => v })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("value was not 0"))

    goodCase and badCase
  }

  def testCollectM = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.collectM("Predicate failed!")({ case v @ 0 => ZIO.succeed(v) })).sandbox.either
    ) must_=== Right(0)

    val partialBadCase = unsafeRun(
      exactlyOnce(0)(_.collectM("Predicate failed!")({ case v @ 0 => ZIO.fail("Partial failed!") })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Partial failed!"))

    val badCase = unsafeRun(
      exactlyOnce(1)(_.collectM("Predicate failed!")({ case v @ 0 => ZIO.succeed(v) })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Predicate failed!"))

    goodCase and partialBadCase and badCase
  }

  def testReject = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.reject({ case v if v != 0 => "Partial failed!" })).sandbox.either
    ) must_=== Right(0)

    val badCase = unsafeRun(
      exactlyOnce(1)(_.reject({ case v if v != 0 => "Partial failed!" })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Partial failed!"))

    goodCase and badCase
  }

  def testRejectM = {
    val goodCase = unsafeRun(
      exactlyOnce(0)(_.rejectM({ case v if v != 0 => ZIO.succeed("Partial failed!") })).sandbox.either
    ) must_=== Right(0)

    val partialBadCase = unsafeRun(
      exactlyOnce(1)(_.rejectM({ case v if v != 0 => ZIO.fail("Partial failed!") })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Partial failed!"))

    val badCase = unsafeRun(
      exactlyOnce(1)(_.rejectM({ case v if v != 0 => ZIO.fail("Partial failed!") })).sandbox.either
    ).left.map(_.failureOrCause) must_=== Left(Left("Partial failed!"))

    goodCase and partialBadCase and badCase
  }

  def testForeachParN_Threads = {
    val n   = 10
    val seq = 0 to 100000
    val res = unsafeRun(IO.foreachParN(n)(seq)(UIO.succeed))
    res must be_===(seq)
  }

  def testForeachParN_Parallel = {
    val io = for {
      p <- Promise.make[Nothing, Unit]
      _ <- UIO.foreachParN(2)(List(UIO.never, p.succeed(())))(a => a).fork
      _ <- p.await
    } yield true
    unsafeRun(io)
  }

  def testForeachParN_Error = {
    val ints = List(1, 2, 3, 4, 5, 6)
    val odds = ZIO.foreachParN(4)(ints) { n =>
      if (n % 2 != 0) ZIO.succeed(n) else ZIO.fail("not odd")
    }
    unsafeRun(odds.either) must_=== Left("not odd")
  }

  def testForeachParN_Interruption = {
    val actions = List(
      ZIO.never,
      ZIO.succeed(1),
      ZIO.fail("C")
    )
    val io = ZIO.foreachParN(4)(actions)(a => a)
    unsafeRun(io.either) must_=== Left("C")
  }
}
